@@ -1,5 +1,5 @@
 /*
- *  linux/drivers/ide/pci/piix.c	Version 0.42	January 11, 2003
+ *  linux/drivers/ide/pci/piix.c	Version 0.44	March 20, 2003
  *
  *  Copyright (C) 1998-1999 Andrzej Krzysztofowicz, Author and Maintainer
  *  Copyright (C) 1998-2000 Andre Hedrick <andre@linux-ide.org>
@@ -103,9 +103,9 @@
 
 #include <asm/io.h>
 
-#include "ide_modes.h"
 #include "piix.h"
 
+static int no_piix_dma;
 #if defined(DISPLAY_PIIX_TIMINGS) && defined(CONFIG_PROC_FS)
 #include <linux/stat.h>
 #include <linux/proc_fs.h>
@@ -114,7 +114,6 @@ static u8 piix_proc = 0;
 #define PIIX_MAX_DEVS		5
 static struct pci_dev *piix_devs[PIIX_MAX_DEVS];
 static int n_piix_devs;
-static int no_piix_dma = 0;
 
 /**
  *	piix_get_info		-	fill in /proc for PIIX ide
@@ -568,7 +567,7 @@ static int piix_config_drive_xfer_rate (ide_drive_t *drive)
 
 	if ((id->capability & 1) && drive->autodma) {
 		/* Consult the list of known "bad" drives */
-		if (hwif->ide_dma_bad_drive(drive))
+		if (__ide_dma_bad_drive(drive))
 			goto fast_ata_pio;
 		if (id->field_valid & 4) {
 			if (id->dma_ultra & hwif->ultra_mask) {
@@ -585,7 +584,7 @@ try_dma_modes:
 				if (!piix_config_drive_for_dma(drive))
 					goto no_dma_set;
 			}
-		} else if (hwif->ide_dma_good_drive(drive) &&
+		} else if (__ide_dma_good_drive(drive) &&
 			   (id->eide_dma_time < 150)) {
 			/* Consult the list of known "good" drives */
 			if (!piix_config_drive_for_dma(drive))
@@ -604,65 +603,6 @@ no_dma_set:
 	return 0;
 }
 
-/**
- *	ich3_busproc		-	bus isolation ioctl
- *	@drive: drive to isolate/restore
- *	@state: bus state to set
- *
- *	Used by the ICH3 to handle bus isolation. We have to do
- *	a little bit of fixing to keep the hardware happy.
- */
- 
-static int ich3_busproc (ide_drive_t * drive, int state)
-{
-	ide_hwif_t *hwif	= HWIF(drive);
-	u32 sig_mode;
-	int shift;
-	int bits;
-	
-	if(hwif->channel == 0)
-		shift = 17;
-	else
-		shift = 19;
-		
-	switch (state) {
-		case BUSSTATE_ON:
-			bits = 0x00;
-			hwif->drives[0].failures = 0;
-			hwif->drives[1].failures = 0;
-			break;
-		case BUSSTATE_OFF:
-			bits = 0x01;
-			break;
-		case BUSSTATE_TRISTATE:
-			bits = 0x10;
-			break;
-		default:
-			return -EINVAL;
-	}
-	
-	if(bits)
-	{
-		int port = hwif->channel == 0 ? 0x40 : 0x42;
-		u16 reg;
-		hwif->drives[0].failures = hwif->drives[0].max_failures + 1;
-		hwif->drives[1].failures = hwif->drives[1].max_failures + 1;
-		/* Turn off IORDY checking to avoid hangs */
-		pci_read_config_word(hwif->pci_dev, port, &reg);
-		reg&=~(1<<5)|(1<<1);
-		pci_write_config_word(hwif->pci_dev, port, reg);
-	}
-	/* Todo: Check locking */
-	pci_read_config_dword(hwif->pci_dev, 0x54, &sig_mode);
-	sig_mode&=~(3<<shift);
-	sig_mode|=(bits<<shift);
-	pci_write_config_dword(hwif->pci_dev, 0x54, sig_mode);
-
-	hwif->bus_state = state;
-	return 0;
-}
-
-
 /**
  *	init_chipset_piix	-	set up the PIIX chipset
  *	@dev: PCI device to set up
@@ -702,7 +642,7 @@ static unsigned int __devinit init_chipset_piix (struct pci_dev *dev, const char
 
 	if (!piix_proc) {
 		piix_proc = 1;
-		ide_pci_register_host_proc(&piix_procs[0]);
+		ide_pci_create_host_proc("piix", piix_get_info);
 	}
 #endif /* DISPLAY_PIIX_TIMINGS && CONFIG_PROC_FS */
 	return 0;
@@ -716,7 +656,7 @@ static unsigned int __devinit init_chipset_piix (struct pci_dev *dev, const char
  *	capabilities of the hardware.
  */
  
-static void __init init_hwif_piix (ide_hwif_t *hwif)
+static void __devinit init_hwif_piix (ide_hwif_t *hwif)
 {
 	u8 reg54h = 0, reg55h = 0, ata66 = 0;
 	u8 mask = hwif->channel ? 0xc0 : 0x30;
@@ -760,10 +700,6 @@ static void __init init_hwif_piix (ide_hwif_t *hwif)
 		case PCI_DEVICE_ID_INTEL_82801AB_1:
 			hwif->ultra_mask = 0x07;
 			break;
-		case PCI_DEVICE_ID_INTEL_82801CA_10:
-		case PCI_DEVICE_ID_INTEL_82801CA_11:
-			hwif->busproc = ich3_busproc;
-			/* fall through */
 		default:
 			pci_read_config_byte(hwif->pci_dev, 0x54, &reg54h);
 			pci_read_config_byte(hwif->pci_dev, 0x55, &reg55h);
@@ -781,23 +717,6 @@ static void __init init_hwif_piix (ide_hwif_t *hwif)
 	hwif->drives[0].autodma = hwif->autodma;
 }
 
-/**
- *	init_dma_piix		-	set up the PIIX DMA
- *	@hwif: IDE interface
- *	@dmabase: DMA PCI base
- *
- *	Set up the DMA on the PIIX controller, providing a DMA base is
- *	available. The PIIX follows the normal specs so we do nothing
- *	magical here.
- */
-
-static void __init init_dma_piix (ide_hwif_t *hwif, unsigned long dmabase)
-{
-	ide_setup_dma(hwif, dmabase, 8);
-}
-
-extern void ide_setup_pci_device(struct pci_dev *, ide_pci_device_t *);
-
 /**
  *	init_setup_piix		-	callback for IDE initialize
  *	@dev: PIIX PCI device
@@ -807,7 +726,7 @@ extern void ide_setup_pci_device(struct pci_dev *, ide_pci_device_t *);
  *	a standard ide PCI setup
  */
  
-static void __init init_setup_piix (struct pci_dev *dev, ide_pci_device_t *d)
+static void __devinit init_setup_piix (struct pci_dev *dev, ide_pci_device_t *d)
 {
 	ide_setup_pci_device(dev, d);
 }
@@ -825,10 +744,11 @@ static int __devinit piix_init_one(struct pci_dev *dev, const struct pci_device_
 {
 	ide_pci_device_t *d = &piix_pci_info[id->driver_data];
 
+	if (!try_module_get(THIS_MODULE))
+		return -ENODEV;
 	if (dev->device != d->device)
 		BUG();
 	d->init_setup(dev, d);
-	MOD_INC_USE_COUNT;
 	return 0;
 }
 
@@ -839,7 +759,7 @@ static int __devinit piix_init_one(struct pci_dev *dev, const struct pci_device_
  *	they are found, disable use of DMA IDE
  */
  
-static void __init piix_check_450nx(void)
+static void __devinit piix_check_450nx(void)
 {
 	struct pci_dev *pdev = NULL;
 	u16 cfg;
@@ -853,8 +773,8 @@ static void __init piix_check_450nx(void)
 		/* Only on the original revision: IDE DMA can hang */
 		if(rev == 0x00)
 			no_piix_dma = 1;
-		/* On all revisions PXB bus lock must be disabled for IDE */
-		else if(cfg & (1<<14))
+		/* On all revisions below 5 PXB bus lock must be disabled for IDE */
+		else if(cfg & (1<<14) && rev < 5)
 			no_piix_dma = 2;
 	}
 	if(no_piix_dma)
@@ -863,7 +783,7 @@ static void __init piix_check_450nx(void)
 		printk(KERN_WARNING "piix: A BIOS update may resolve this.\n");
 }		
 
-static struct pci_device_id piix_pci_tbl[] __devinitdata = {
+static struct pci_device_id piix_pci_tbl[] = {
 	{ PCI_VENDOR_ID_INTEL, PCI_DEVICE_ID_INTEL_82371FB_0, PCI_ANY_ID, PCI_ANY_ID, 0, 0, 0},
 	{ PCI_VENDOR_ID_INTEL, PCI_DEVICE_ID_INTEL_82371FB_1, PCI_ANY_ID, PCI_ANY_ID, 0, 0, 1},
 	{ PCI_VENDOR_ID_INTEL, PCI_DEVICE_ID_INTEL_82371MX,   PCI_ANY_ID, PCI_ANY_ID, 0, 0, 2},
@@ -882,11 +802,14 @@ static struct pci_device_id piix_pci_tbl[] __devinitdata = {
 	{ PCI_VENDOR_ID_INTEL, PCI_DEVICE_ID_INTEL_82801EB_11,PCI_ANY_ID, PCI_ANY_ID, 0, 0, 15},
 	{ PCI_VENDOR_ID_INTEL, PCI_DEVICE_ID_INTEL_82801E_11, PCI_ANY_ID, PCI_ANY_ID, 0, 0, 16},
 	{ PCI_VENDOR_ID_INTEL, PCI_DEVICE_ID_INTEL_82801DB_10,PCI_ANY_ID, PCI_ANY_ID, 0, 0, 17},
+#ifndef CONFIG_SCSI_SATA
  	{ PCI_VENDOR_ID_INTEL, PCI_DEVICE_ID_INTEL_82801EB_1, PCI_ANY_ID, PCI_ANY_ID, 0, 0, 18},
+#endif /* !CONFIG_SCSI_SATA */
 	{ PCI_VENDOR_ID_INTEL, PCI_DEVICE_ID_INTEL_ESB_2, PCI_ANY_ID, PCI_ANY_ID, 0, 0, 19},
 	{ PCI_VENDOR_ID_INTEL, PCI_DEVICE_ID_INTEL_ICH6_2, PCI_ANY_ID, PCI_ANY_ID, 0, 0, 20},
 	{ 0, },
 };
+MODULE_DEVICE_TABLE(pci, piix_pci_tbl);
 
 static struct pci_driver driver = {
 	.name		= "PIIX IDE",
@@ -894,13 +817,13 @@ static struct pci_driver driver = {
 	.probe		= piix_init_one,
 };
 
-static int piix_ide_init(void)
+static int __init piix_ide_init(void)
 {
 	piix_check_450nx();
 	return ide_pci_register_driver(&driver);
 }
 
-static void piix_ide_exit(void)
+static void __exit piix_ide_exit(void)
 {
 	ide_pci_unregister_driver(&driver);
 }
@@ -911,5 +834,3 @@ module_exit(piix_ide_exit);
 MODULE_AUTHOR("Andre Hedrick, Andrzej Krzysztofowicz");
 MODULE_DESCRIPTION("PCI driver module for Intel PIIX IDE");
 MODULE_LICENSE("GPL");
-
-EXPORT_NO_SYMBOLS;
